/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/io.h>

#include <locale.h>

#include <hlog/country.h>
#include <hlog/index.h>
#include <hlog/version.h>

#include "wsjtx.h"

#define DATA_DIR_CONFIG_FNAME	"config.scm"
#define DATA_DIR_COUNTRY_FNAME	"cty.csv"

static int data_dir_fd;

static int load_countries(void)
{
	int ret;

	printf("Loading countries data...\n");
	ret = country_load(data_dir_fd, DATA_DIR_COUNTRY_FNAME);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load country data: %s\n",
			xstrerror(ret));

	return ret;
}

static int load_index(void)
{
	int ret;

	printf("Loading existing QSOs...\n");
	ret = index_init(data_dir_fd);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load existing entries: %s\n",
			xstrerror(ret));

	return ret;
}

static void usage(const char *prog)
{
	fprintf(stderr, "HLog version %s (libjeffpc %s)\n\n", version_string,
		jeffpc_version);

	fprintf(stderr, "Usage: %s <path to data dir>\n", prog);
	exit(1);
}

int main(int argc, char **argv)
{
	const char *data_dir;
	int ret;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	if (argc != 2)
		usage(argv[0]);

	data_dir = argv[1];
	data_dir_fd = xopen(data_dir, O_RDONLY, 0);
	if (data_dir_fd < 0) {
		fprintf(stderr, "Error: failed to open data directory: %s\n",
			xstrerror(data_dir_fd));
		goto err;
	}

	ret = load_countries();
	if (ret)
		goto err_close;

	ret = load_index();
	if (ret)
		goto err_close;

	ret = wsjtx_start(data_dir_fd);
	if (ret)
		goto err_close;

	for (;;)
		sleep(60);

	xclose(data_dir_fd);

	return 0;

err_close:
	xclose(data_dir_fd);

err:
	return 1;
}
